/*
   clock.c

   provides a simulated clock interrupt, and routines
   for manipulating them in Windows NT on x86.

   You should not need to modify this file.
*/

#include <stdio.h>
#include "clock.h"
#include "assert.h"
#include "minithread_md.h"
#include "minithread_public.h"
#include "network.h"

#define DEBUG 0

/* get the correct version of windows nt recognised */
#define _WIN32_WINNT 0x0400
#include <windows.h>
#include <winsock.h>

typedef struct signal_queue_t signal_queue_t;
struct signal_queue_t {
  HANDLE threadid;
  CONTEXT *context;
  int intnumber;
  signal_queue_t *next;
};

/* threads which have finished servicing their interrupts */ 
static signal_queue_t *signalq = NULL;

/* return thread waits on this semaphore to do an interrupt return */
static HANDLE cleanup = NULL;

/* handle that protects the globals in this package, including
   the queue of threads that are handling simulated interrupts,
   as well as the main threads stack */
static HANDLE mutex = NULL;

/* categories of interrupt types */
typedef enum { CLOCK_INTERRUPT, NETWORK_INTERRUPT } interrupt_type_t;


typedef struct {
  network_address_t addr;
  char buffer[MAX_NETWORK_PKT_SIZE];
  int size;
} network_interrupt_arg_t;

typedef network_interrupt_arg_t interrupt_arg_t;


/* Are interrupts enabled? The interrupt_handler function be called
   iff this is true. */

interrupt_level_t interrupt_level;

/* global handle to the user's clock interrupt service routine */
static clock_handler_t clock_handler = NULL;

/* global handle to the user's network interrupt service routine */
static network_handler_t network_handler = NULL;

static HANDLE clock_thread;   /* NT thread to check for timer events */
static HANDLE system_thread;  /* NT thread for running the minithreads */
static HANDLE return_thread;  /* NT thread for restarting system thread */

static int pid;

/* is the main thread ready to be teleported back to */
/* previous state by the interrupt return assist thread ? */
static int readytoreturn = 0;

interrupt_level_t set_interrupt_level(interrupt_level_t newlevel) {
  return swap(&interrupt_level, newlevel);
}

void loopforever() {
  for(;;)
    ;
  /* NOT REACHED */
  abort();
}

void loopforever_end() {
}

/* run the user's supplied interrupt handler and clean up the state after it:
   we wake up the return thread and then loop indefinitely; it overwrites our
   execution context to set us back to where we were before the interrupt
	 if the argument "arg" is NULL, it's a clock interrupt; otherwise, it's a
	 pointer to a network_interrupt_arg_t, which contains information about the
	 packet which has arrived.

   disables interrupts before passing control to the return thread.
*/
void __fastcall receive_interrupt(CONTEXT* context, interrupt_arg_t* arg) {
  signal_queue_t* sq;
  
	readytoreturn = FALSE;

  if (DEBUG)
    printf("SYS:Running user interrupt handler, context = 0x%x, arg = 0x%x.\n",
	   context, arg);
  
  /* now, call the appropriate interrupt handler */
  if (arg == NULL) {
    if (DEBUG)
      printf("SYS:Clock interrupt.\n");
    if (clock_handler != NULL)
      clock_handler();
  }
	else {
    network_interrupt_arg_t* packet = (network_interrupt_arg_t *) arg;
    if (DEBUG) {
      printf("SYS:Network interrupt.\n");
    }
    if (network_handler != NULL)
      network_handler(packet->addr, packet->buffer, packet->size);
    free(packet);
  }
  
  if (WaitForSingleObject(mutex, INFINITE) != WAIT_OBJECT_0) {
		printf("Error: code %d.\n", GetLastError());
		assert(0);
	}
  assert(signalq == NULL);

  sq = (signal_queue_t *) malloc(sizeof(signal_queue_t));
  sq->context = context;
  sq->threadid = system_thread;

  sq->next = signalq;
  signalq = sq;

  if (interrupt_level == ENABLED) {
    if (DEBUG)
      printf("SYS:disabling interrupts in handler.\n");
    interrupt_level = DISABLED;
  }  
  ReleaseMutex(mutex);

  /* this is correct but not elegant, we need an atomic way to wake up the
     assist thread and go to sleep ourselves. One could use SignalAndWait
     here, but it does not work on CE */
  ReleaseSemaphore(cleanup, 1, NULL); /* tell the return thread to run */

  readytoreturn = TRUE;

	/*  
  if (DEBUG)
    printf("SYS:Looping forever.\n");
	*/
  loopforever();
}

/* this thread assists with the returns from simulated interrupts

   ideally, the interrupthandler would just assume the state right before the
   interrupt, but this is not possible given the x86 and nt, so the interrupt
   handling thread has to suspend itself, and let some other thread manipulate
   its state while it is suspended. 

   when this thread runs, interrupts will be disabled, so we reenable them
   once it is finished.
*/
int WINAPI interrupt_return_assist(void* ptr) {
  signal_queue_t *sq;
  int scount, rcount;
  int success = FALSE;
  CONTEXT context;

  if (DEBUG)
    printf("IRA:starting ...\n");

  for(;;) {
    /* wait politely until we are needed */
    if (WaitForSingleObject(cleanup, INFINITE) != WAIT_OBJECT_0) {
			printf("Error: code %d.\n", GetLastError());
			assert(0);
		}

		if (DEBUG)
			printf("IRA:woken up ...\n");
	
		if (WaitForSingleObject(mutex, INFINITE) != WAIT_OBJECT_0) {
			printf("Error: code %d.\n", GetLastError());
			assert(0);
		}
		{
			success = FALSE;
			sq = signalq;
			signalq = signalq->next;
			assert(signalq == NULL);

			/* wait for the system thread to enter a "safe" state */
			while(readytoreturn != TRUE)
				SwitchToThread();

      while(success != TRUE) {

				scount = SuspendThread(sq->threadid);
				memset(&context, 0, sizeof(CONTEXT));
				context.ContextFlags = 
					CONTEXT_FULL | CONTEXT_FLOATING_POINT | CONTEXT_DEBUG_REGISTERS;
				assert(GetThreadContext(system_thread, &context) != 0);
				if (DEBUG)
					printf("IRA:system thread is at 0x%x.\n", context.Eip);
				if((context.Eip >= (unsigned int)loopforever && 
					  context.Eip <= (unsigned int)loopforever_end)) {

					if (DEBUG)
						printf("IRA:enabling interrupts.\n");
					interrupt_level = ENABLED;

					assert(SetThreadContext(sq->threadid, sq->context) != 0);
					rcount = ResumeThread(sq->threadid);

					if (DEBUG)
						printf("IRA:Interrupt return assist scount %d rcount %d\n", 
									 scount, rcount);

					assert(rcount >= scount + 1);
					success = TRUE;
				} 
				else {
					ResumeThread(system_thread);
					ReleaseMutex(mutex);
					SwitchToThread();
					if (WaitForSingleObject(mutex, INFINITE) != WAIT_OBJECT_0) {
						printf("Error: code %d.\n", GetLastError());
						assert(0);
					}
				}
      }
    } 
    ReleaseMutex(mutex);
    free(sq);
  }

  /* never reached */
  return 0;
}

/* send an interrupt to the system thread: adjust its stack to call the 
	 appropriate interrupt handler with the specified argument. the "type"
	 argument identifies interrupt-specific processing which must be done,
	 in particular, what we should do if interrupts are disabled or the
	 system thread is in a non-preemptable state (e.g. executing a system
	 library function). clock interrupts are dropped, network interrupts
	 are deferred. this function replaces code which used to be common
	 to clock_poll and network_poll.
*/
void send_interrupt(interrupt_type_t type, interrupt_arg_t* arg) {
  CONTEXT context;
	int safe_to_proceed = 0;
	int drop_interrupt = 0;

	for (;;) {
		if (WaitForSingleObject(mutex, INFINITE) != WAIT_OBJECT_0) {
			printf("Error: code %d.\n", GetLastError());
			assert(0);
		}

		/* need to protect this code: we only activate the interrupt if interrupts
			 are actually enabled, but we also need to ensure they are not disabled
			 after we test -- so we suspend the system thread first.
		*/
		SuspendThread(system_thread);
		memset(&context, 0, sizeof(CONTEXT));
		context.ContextFlags = 
			CONTEXT_FULL | CONTEXT_FLOATING_POINT | CONTEXT_DEBUG_REGISTERS;
		assert(GetThreadContext(system_thread, &context) != 0);

		/* interrupt-type-specific processing: can we interrupt now? */
		switch (type) {
		case CLOCK_INTERRUPT:			
			if (interrupt_level == DISABLED
				  || (context.Eip < (unsigned int) start)
					|| (context.Eip > (unsigned int) end)) {
				drop_interrupt = 1;
			}
			else
				safe_to_proceed = 1;
			break;
		case NETWORK_INTERRUPT:
			if (interrupt_level == ENABLED 
				  && (context.Eip >= (unsigned int) start)
					&& (context.Eip <= (unsigned int) end)) {
				interrupt_level = DISABLED;
				safe_to_proceed = 1;
			}
			break;
		default:
			break;
		}

		if (safe_to_proceed == 1)
			break;
 		else {
			ResumeThread(system_thread);
			ReleaseMutex(mutex);
		}

		if (drop_interrupt == 1)
			return;
		else
			SwitchToThread();
	}

	/* now fix the system thread's stack so it runs run_user_handler */
  {
    int stack;
    int oldpc = 0;

    if (DEBUG) {
      printf("Interrupts are enabled, system thread pc = 0x%x\n", context.Eip);
      printf("Adjusting system thread context ...\n");
    }

    /* set the interrupt number */
    /* arg->intnumber = ++intnumber; */

    oldpc = context.Eip;
    stack = context.Esp; /* Esp == extended stack pointer */

    /* safe to do a printf because other thread is stunned in user code */
    if (DEBUG)
      printf("Suspended system thread, adjusting stack, sp = 0x%x\n", stack);

    stack -= (sizeof(CONTEXT) + 64); /* 64 is slop */
    memcpy((int *) stack, &context, sizeof(CONTEXT));
    context.Esp = stack;
    context.Eip = (int) ((void *) receive_interrupt);
    context.Edx = (int) arg;
    context.Ecx = stack;
    
    assert(SetThreadContext(system_thread, &context) != 0);
    assert(GetThreadContext(system_thread, &context) != 0);
    ResumeThread(system_thread);
  }
  ReleaseMutex(mutex);
}

/* procedure to poll the event queue for timer events, run by the clock
   thread; on a timer event, call "send_interrupt()" to run the system thread's
   clock handler routine 
*/
int WINAPI clock_poll(void* arg) {
  LARGE_INTEGER i;
  HANDLE timer; 
  HANDLE thread = GetCurrentThread();
  char name[32];

	sprintf(name, "timer %d", pid);
	timer = CreateWaitableTimer(NULL, TRUE, name);
  assert(timer != NULL);

  for (;;) {
    i.QuadPart = -PERIOD*10; /* NT timer values are in hundreds of nanoseconds */
    assert(SetWaitableTimer(timer, &i, 0, NULL, NULL, FALSE) != 0);
    if (WaitForSingleObject(timer, INFINITE) == WAIT_OBJECT_0) {
			if (DEBUG)
				printf("CLK: clock tick.\n");
      send_interrupt(CLOCK_INTERRUPT, NULL);
		}
  }

  /* never reached */
  return 0;
}

/* receive incoming packets from the specified socket, translate their 
   addresses to network_address_t type, and call the user-supplied handler

   a network interrupt disables interrupts, so that we can cleanly return.
   if interrupts were not disabled, and we were hit by a clock interrupt, it
   would not be possible to return until we returned to the original stack.
   this is inelegant, but we are constrained by ignorance of the minithread
   struct format!
*/
int WINAPI network_poll(void* arg) {
  SOCKET* s;
  network_interrupt_arg_t* packet;
  struct sockaddr_in addr;
  int fromlen = sizeof(struct sockaddr_in);

  s = (SOCKET *) arg;

  for (;;) {

    /* we rely on run_user_handler to destroy this data structure */
    if (DEBUG)
      printf("NET:Allocating an incoming packet.\n");

    packet = 
      (network_interrupt_arg_t *) malloc(sizeof(network_interrupt_arg_t));
    assert(packet != NULL);
  
    packet->size = recvfrom(*s, packet->buffer, MAX_NETWORK_PKT_SIZE,
			    0, (struct sockaddr *) &addr, &fromlen);
    if (packet->size <= 0)
      printf("NET:Error, %d.\n", WSAGetLastError());
    else if (DEBUG)
      printf("NET:Received a packet, seqno %d.\n", ntohl(*((int *) packet->buffer)));
   
    assert(packet->size > 0);
    assert(fromlen == sizeof(struct sockaddr_in));
    sockaddr_to_network_address(&addr, packet->addr);

    /* now we have filled in the arg to the network interrupt service routine,
       so we have to get the user's thread to run it.
    */
		if (DEBUG)
			printf("NET:packet arrived.\n");
		send_interrupt(NETWORK_INTERRUPT, packet);
	}	
}

/* start polling for network packets. this is separate so that clock interrupts
   can be turned on without network interrupts. however, this function requires
   that clock_init has been called! */
void start_network_poll(network_handler_t h, SOCKET* s) {
  HANDLE network_thread = NULL; /* NT thread to check for incoming packets */
  int id;

  network_handler = h;

  printf("Starting network interrupts.\n");
  
  /* create clock and return threads, but discard ids */
  network_thread = CreateThread(NULL, 0, network_poll, s, 0, &id); 
  assert(network_thread != NULL);
}

/* setup the interval timer and install user interrupt handler.  After this
   routine is called, and after you call set_interrupt_level(ENABLED), clock
   interrupts will begin to be sent.  They will call the handler
   function h specified by the caller.
*/
void minithread_clock_init(clock_handler_t h)
{
  HANDLE process;
  int id;
  char name[32];
	clock_handler = h;

  if (h == NULL) {
    printf("Must provide an interrupt handler, interrupts not started.\n");
    return;
  }

  printf("Starting clock interrupts.\n");

	pid = GetCurrentProcessId();

  /* set up the signal queue and semaphores */
	sprintf(name, "interrupt mutex %d", pid);
  mutex = CreateMutex(NULL, FALSE, name);

  /* use this semaphore as a condition variable */
	sprintf(name, "interrupt return semaphore %d", pid);
  cleanup = CreateSemaphore(NULL, 0, 10, name);

  interrupt_level = DISABLED;

  /* overcome an NT "feature" -- GetCurrentThread() returns a "pseudohandle" to
     the executing thread, not the real handle, so we need to use 
     DuplicateHandle() to get the real handle.
  */
  process = GetCurrentProcess();
  assert(DuplicateHandle(process, GetCurrentThread(), process, 
			 &system_thread, THREAD_ALL_ACCESS, FALSE, 0) == TRUE);

  /* create clock and return threads, but discard ids */
  clock_thread = CreateThread(NULL, 0, clock_poll, NULL, 0, &id); 
  assert(clock_thread != NULL);

  return_thread = CreateThread(NULL, 0, interrupt_return_assist, NULL, 0, &id);
  assert(return_thread != NULL);
}
