/*
 * network.c:
 *	This module paints the unix socket interface a pretty color.
 */


#define _WIN32_WINNT 0x0400
#include <windows.h>
#include <winsock.h>

#include "network.h"
#include "clock.h"

#include <assert.h>
#include <string.h>

#include <sys/types.h>
#include <stdio.h>
#include <fcntl.h>
#include "minithread.h"
#include "minithread_public.h"

short my_udp_port = MINIMSG_PORT;
short other_udp_port = MINIMSG_PORT;

double loss_rate = 0.0;
double duplication_rate = 0.0;
int synthetic_network = 0;

WSADATA winsock_version_data;

struct address_info {
  SOCKET sock;
  struct sockaddr_in sin;
  char pkt[MAX_NETWORK_PKT_SIZE];
};

struct address_info if_info;
static tas_lock_t initialized = 0;


void 
network_address_to_sockaddr(network_address_t addr, struct sockaddr_in *sin);

void 
sockaddr_to_network_address(struct sockaddr_in *sin, network_address_t addr);

int
network_initialize(network_handler_t handler) {
  int arg = 1;

  /* initialise the NT socket library, inexplicably required by NT */
  assert(WSAStartup(MAKEWORD(2, 0), &winsock_version_data) == 0);
  
  if (atomic_test_and_set(&initialized)) {
    return -1;
  }

  memset(&if_info, 0, sizeof(if_info));
  
  if_info.sock = socket(PF_INET, SOCK_DGRAM, 0);
  if (if_info.sock < 0)  {
    perror("socket");
    return -1;
  }

  if_info.sin.sin_family = SOCK_DGRAM;
  if_info.sin.sin_addr.s_addr = htonl(0);
  if_info.sin.sin_port = htons(my_udp_port);
  if (bind(if_info.sock, (struct sockaddr *) &if_info.sin, 
	   sizeof(if_info.sin)) < 0)  {
    printf("Error: code %d.\n", GetLastError());
    return -1;
  }

  /* set for fast reuse */
  assert(setsockopt(if_info.sock, SOL_SOCKET, SO_REUSEADDR, 
		    (char *) &arg, sizeof(int)) == 0);

  /*
   * Interrupts are handled through the caller's handler.
   */
  
  start_network_poll(handler, &if_info.sock);
    
  return 0;
}

static int
send_pkt(network_address_t dest_address, 
		 int hdr_len, char* hdr, 
		 int data_len, char* data) {
  int cc;
  struct sockaddr_in sin;
  char* bufp;
  int sz, pktlen;

  pktlen = hdr_len + data_len;

  /* sanity checks */
  if (hdr_len < 0 || data_len < 0 || pktlen > MAX_NETWORK_PKT_SIZE)
    return 0;
  
  /*
   * Pull up the headers and data and stuff them into the output
   * packet with the
   * field sizes embedded.
   */
  
  bufp = if_info.pkt;

  sz = hdr_len;
  memcpy(bufp, hdr, sz);
  bufp += sz;

  sz = data_len;
  memcpy(bufp, data, sz);
  bufp += sz;
  
  network_address_to_sockaddr(dest_address, &sin);
  cc = sendto(if_info.sock,
	      if_info.pkt,
	      pktlen,
	      0,
	      (struct sockaddr *) &sin,
	      sizeof(sin));

  return cc;
}

int
network_send_pkt(network_address_t dest_address, int hdr_len, char* hdr, int data_len, char* data) {

	if (synthetic_network) {
	  if((rand()*1.0)/(RAND_MAX*1.0) < loss_rate)
		  return (hdr_len+data_len);

  	if((rand()*1.0)/(RAND_MAX*1.0) < duplication_rate)
			send_pkt(dest_address, hdr_len, hdr, data_len, data);
	}

	return send_pkt(dest_address, hdr_len, hdr, data_len, data);
}

void
network_get_my_address(network_address_t my_address) {
	char hostname[64];
	assert(gethostname(hostname, 64) == 0);
  network_translate_hostname(hostname, my_address);
	my_address[1] = (long) ntohs(my_udp_port);
}

int
network_translate_hostname(char* hostname, network_address_t address) {
  struct hostent* host = gethostbyname(hostname);
  if (host == NULL)
    return -1;
  else {
    address[0] = (long) *((int *) host->h_addr);
    address[1] = (long) ntohs(other_udp_port);
    return 0;
  }
}

void
sockaddr_to_network_address(struct sockaddr_in* sin, network_address_t addr) {
  addr[0] = sin->sin_addr.s_addr;
  addr[1] = sin->sin_port;
}

void
network_address_to_sockaddr(network_address_t addr, struct sockaddr_in* sin) {
  memset(sin, 0, sizeof(*sin));
  sin->sin_addr.s_addr = addr[0];
  sin->sin_port = (short)addr[1];
  sin->sin_family = SOCK_DGRAM;
}

void
network_udp_ports(short myportnum, short otherportnum) {
	my_udp_port = myportnum;
	other_udp_port = otherportnum;
}

void
network_synthetic_params(double loss, double duplication) {
	synthetic_network = 1;	
	loss_rate = loss;
	duplication_rate = duplication;
}